@@ -92,6 +92,7 @@ typedef struct {
 
 static GsfInput *gsf_infile_msole_new_child (GsfInfileMSOle *parent,
 					     MSOleDirent *dirent);
+static void ole_info_unref (MSOleInfo *info);
 
 /**
  * ole_get_block :
@@ -216,6 +217,9 @@ ole_info_get_sb_file (GsfInfileMSOle *parent)
 	parent->info->sb_file = gsf_infile_msole_new_child (parent,
 		parent->info->root_dir);
 
+	/* avoid creating a circular reference */
+	ole_info_unref (((GsfInfileMSOle *)parent->info->sb_file)->info);
+
 	g_return_val_if_fail (parent->info->sb.bat.block == NULL, NULL);
 
 	if (ole_make_bat (&parent->info->bb.bat,
@@ -297,6 +301,7 @@ ole_dirent_new (GsfInfileMSOle *ole, guint32 entry, MSOleDirent *parent)
 	next  = GSF_LE_GET_GUINT32 (data + DIRENT_NEXT);
 	child = GSF_LE_GET_GUINT32 (data + DIRENT_CHILD);
 	name_len = GSF_LE_GET_GUINT16 (data + DIRENT_NAME_LEN);
+	dirent->name = NULL;
 	if (0 < name_len && name_len <= DIRENT_MAX_NAME_SIZE) {
 		gunichar2 uni_name [DIRENT_MAX_NAME_SIZE];
 		guint8 const *end;
@@ -315,7 +320,9 @@ ole_dirent_new (GsfInfileMSOle *ole, guint32 entry, MSOleDirent *parent)
 			dirent->name = g_utf16_to_utf8 (uni_name, -1, NULL, NULL, NULL);
 		} else
 			dirent->name = g_strndup ((gchar *)data, (gsize)(end - data + 1));
-	} else
+	}
+	/* be really anal in the face of screwups */
+	if (dirent->name == NULL)
 		dirent->name = g_strdup ("");
 	dirent->collation_name = g_utf8_collate_key (dirent->name, -1);
 
@@ -363,6 +370,7 @@ ole_dirent_free (MSOleDirent *dirent)
 }
 
 /*****************************************************************************/
+
 static void
 ole_info_unref (MSOleInfo *info)
 {
@@ -569,12 +577,15 @@ gsf_infile_msole_finalize (GObject *obj)
 		g_object_unref (G_OBJECT (ole->input));
 		ole->input = NULL;
 	}
-	if (ole->info != NULL) {
+	if (ole->info != NULL &&
+	    ole->info->sb_file != (GsfInput *)ole) {
 		ole_info_unref (ole->info);
 		ole->info = NULL;
 	}
 	ols_bat_release (&ole->bat);
 
+	g_free (ole->stream.buf);
+
 	parent_class = g_type_class_peek (GSF_INFILE_TYPE);
 	if (parent_class && parent_class->finalize)
 		parent_class->finalize (obj);
