@@ -116,23 +116,24 @@ gsf_outfile_msole_seek (GsfOutput *output, gsf_off_t offset,
 	switch (whence) {
 	case G_SEEK_SET : break;
 	case G_SEEK_CUR : offset += output->cur_offset;	break;
-	case G_SEEK_END : offset += output->cur_size;		break;
-	default : return TRUE;
+	case G_SEEK_END : offset += output->cur_size;	break;
+	default :
+		break; /*checked in GsfOutput wrapper */
 	}
 
 	switch (ole->type) {
 	case MSOLE_DIR:
 		if (offset != 0) {
 			g_warning ("Attempt to seek a directory");
-			return TRUE;
+			return FALSE;
 		}
-		return FALSE;
+		return TRUE;
 
 	case MSOLE_SMALL_BLOCK:
 		/* it is ok to seek past the big block threshold
 		 * we don't convert until they _write_ something
 		 */
-		return FALSE;
+		return TRUE;
 
 	case MSOLE_BIG_BLOCK:
 		return gsf_output_seek (ole->sink,
@@ -140,10 +141,10 @@ gsf_outfile_msole_seek (GsfOutput *output, gsf_off_t offset,
 			G_SEEK_SET);
 
 	default :
-		return TRUE;
+		return FALSE;
 	}
 
-	return TRUE;
+	return FALSE;
 }
 
 static void
@@ -515,6 +516,31 @@ gsf_outfile_msole_write (GsfOutput *output,
 }
 
 
+#define GET_OUTPUT_CLASS(instance) \
+         G_TYPE_INSTANCE_GET_CLASS (instance, GSF_OUTPUT_TYPE, GsfOutputClass)
+
+static gboolean
+gsf_outfile_msole_vprintf (GsfOutput *output, char const *format, va_list args)
+{
+	GsfOutfileMSOle *ole = (GsfOutfileMSOle *)output;
+	GsfOutputClass *klass;
+	int res;
+
+	g_return_val_if_fail (ole->type != MSOLE_DIR, FALSE);
+
+	if (ole->type == MSOLE_BIG_BLOCK) {
+		klass = GET_OUTPUT_CLASS (ole->sink);
+		res = klass->Vprintf (ole->sink, format, args);
+	} else {
+		klass = (GsfOutputClass *) (g_type_class_peek_parent
+					    (GET_OUTPUT_CLASS (output)));
+		res = klass->Vprintf (output, format, args);
+	}
+
+	return res;
+}
+
+
 static void
 ole_register_child (GsfOutfileMSOle *root, GsfOutfileMSOle *child)
 {
@@ -594,13 +620,14 @@ gsf_outfile_msole_init (GObject *obj)
 static void
 gsf_outfile_msole_class_init (GObjectClass *gobject_class)
 {
-	GsfOutputClass  *input_class  = GSF_OUTPUT_CLASS (gobject_class);
+	GsfOutputClass  *output_class  = GSF_OUTPUT_CLASS (gobject_class);
 	GsfOutfileClass *outfile_class = GSF_OUTFILE_CLASS (gobject_class);
 
 	gobject_class->finalize		= gsf_outfile_msole_finalize;
-	input_class->Close		= gsf_outfile_msole_close;
-	input_class->Seek		= gsf_outfile_msole_seek;
-	input_class->Write		= gsf_outfile_msole_write;
+	output_class->Close		= gsf_outfile_msole_close;
+	output_class->Seek		= gsf_outfile_msole_seek;
+	output_class->Write		= gsf_outfile_msole_write;
+	output_class->Vprintf		= gsf_outfile_msole_vprintf;
 	outfile_class->new_child	= gsf_outfile_msole_new_child;
 }
 
