@@ -41,7 +41,7 @@ struct _GsfOutfileZip {
 
 	ZipVDir       *vdir;
 	GPtrArray     *root_order;	/* only valid for the root */
-	
+
 	z_stream  *stream;
 	GsfZipCompressionMethod compression_method;
 
@@ -58,6 +58,25 @@ typedef struct {
 #define GSF_OUTFILE_ZIP_CLASS(k)    (G_TYPE_CHECK_CLASS_CAST ((k), GSF_OUTFILE_ZIP_TYPE, GsfOutfileZipClass))
 #define GSF_IS_OUTFILE_ZIP_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GSF_OUTFILE_ZIP_TYPE))
 
+
+static void
+disconnect_children (GsfOutfileZip *zip)
+{
+	unsigned i;
+
+	if (!zip->root_order)
+		return;
+
+	for (i = 0 ; i < zip->root_order->len ; i++) {
+		GsfOutfileZip *child =
+			g_ptr_array_index (zip->root_order, i);
+		if (child)
+			g_object_unref (G_OBJECT (child));
+	}
+	g_ptr_array_free (zip->root_order, FALSE);
+	zip->root_order = NULL;
+}
+
 static void
 gsf_outfile_zip_finalize (GObject *obj)
 {
@@ -66,9 +85,13 @@ gsf_outfile_zip_finalize (GObject *obj)
 
 	GsfOutput *output = (GsfOutput *)obj;
 
-	if (!gsf_output_is_closed (output))
+	if (!gsf_output_is_closed (output)) {
 		gsf_output_close (output);
 
+		/* If the closing failed, we might have stuff here.  */
+		disconnect_children (zip);
+	}
+
 	if (zip->sink != NULL) {
 		g_object_unref (G_OBJECT (zip->sink));
 		zip->sink = NULL;
@@ -78,9 +101,10 @@ gsf_outfile_zip_finalize (GObject *obj)
 		(void) deflateEnd (zip->stream);
 	g_free (zip->stream);
 	g_free (zip->buf);
-	
-	vdir_free (zip->vdir, TRUE); /* Frees vdirs recursively */
-	
+
+	if (zip == zip->root)
+		vdir_free (zip->vdir, TRUE); /* Frees vdirs recursively */
+
 	parent_class = g_type_class_peek (GSF_OUTFILE_TYPE);
 	if (parent_class && parent_class->finalize)
 		parent_class->finalize (obj);
@@ -92,7 +116,7 @@ gsf_outfile_zip_seek (GsfOutput *output, gsf_off_t offset, GSeekType whence)
 	(void) output;
 	(void) offset;
 	(void) whence;
-	
+
 	return FALSE;
 }
 
@@ -104,7 +128,7 @@ zip_dirent_write (GsfOutput *sink, ZipDirent *dirent)
 	guint8 buf[ZIP_DIRENT_SIZE];
 	int nlen = strlen (dirent->name);
 	gboolean ret;
-	
+
 	memset (buf, 0, sizeof buf);
 	memcpy (buf, dirent_signature, sizeof dirent_signature);
 	GSF_LE_SET_GUINT16 (buf + ZIP_DIRENT_ENCODER, 0x317); /* Unix */
@@ -122,9 +146,9 @@ zip_dirent_write (GsfOutput *sink, ZipDirent *dirent)
 	GSF_LE_SET_GUINT16 (buf + ZIP_DIRENT_DISKSTART, 0);
 	GSF_LE_SET_GUINT16 (buf + ZIP_DIRENT_FILE_TYPE, 0);
 	/* Hardcode file mode 644 */
-	GSF_LE_SET_GUINT32 (buf + ZIP_DIRENT_FILE_MODE, 0644 << 16); 
+	GSF_LE_SET_GUINT32 (buf + ZIP_DIRENT_FILE_MODE, 0644 << 16);
 	GSF_LE_SET_GUINT32 (buf + ZIP_DIRENT_OFFSET, dirent->offset);
-	
+
 	ret = gsf_output_write (sink, sizeof buf, buf);
 	if (ret)
 		ret = gsf_output_write (sink, nlen, dirent->name);
@@ -133,13 +157,12 @@ zip_dirent_write (GsfOutput *sink, ZipDirent *dirent)
 }
 
 static gboolean
-zip_trailer_write (GsfOutfileZip *zip, gsf_off_t dirpos)
+zip_trailer_write (GsfOutfileZip *zip, unsigned entries, gsf_off_t dirpos)
 {
 	static guint8 const trailer_signature[] =
 		{ 'P', 'K', 0x05, 0x06 };
 	guint8 buf[ZIP_TRAILER_SIZE];
 	gsf_off_t pos = gsf_output_tell (zip->sink);
-	unsigned entries = zip->root_order->len;
 
 	memset (buf, 0, sizeof buf);
 	memcpy (buf, trailer_signature, sizeof trailer_signature);
@@ -154,41 +177,53 @@ zip_trailer_write (GsfOutfileZip *zip, gsf_off_t dirpos)
 static gboolean
 zip_close_root (GsfOutput *output)
 {
-	GsfOutfileZip *zip = GSF_OUTFILE_ZIP (output);	
+	GsfOutfileZip *zip = GSF_OUTFILE_ZIP (output);
 	GsfOutfileZip *child;
 	gsf_off_t dirpos = gsf_output_tell (zip->sink);
 	GPtrArray *elem = zip->root_order;
+	unsigned entries = elem->len;
 	unsigned i;
 
 	/* Check that children are closed */
 	for (i = 0 ; i < elem->len ; i++) {
 		child = g_ptr_array_index (elem, i);
 		if (!gsf_output_is_closed (GSF_OUTPUT (child))) {
-			g_message ("Child still open");
+			g_warning ("Child still open");
 			return FALSE;
 		}
 	}
+
 	/* Write directory */
-	for (i = 0 ; i < elem->len ; i++) {
+	for (i = 0 ; i < entries ; i++) {
 		child = g_ptr_array_index (elem, i);
 		if (!zip_dirent_write (zip->sink, child->vdir->dirent))
 			return FALSE;
-	}		
+	}
 
-	return zip_trailer_write (zip, dirpos);
+	disconnect_children (zip);
+
+	return zip_trailer_write (zip, entries, dirpos);
 }
 
 static int
-stream_name_len (GsfOutfile *outfile)
+stream_name_len (GsfOutfileZip *zip)
 {
-	GsfOutput  *output = GSF_OUTPUT (outfile);
-	GsfOutfile *container = gsf_output_container (output);
-	char const *name = gsf_output_name (output);
-	int len = name ? strlen (name) : 0;
+	GsfOutput  *output;
+	GsfOutfile *container;
+	char const *name;
+	int len = 0;
 	int cnlen;
 
+	if (zip == zip->root)
+		return 0;
+
+	output = GSF_OUTPUT (zip);
+	container = gsf_output_container (output);
+	name = gsf_output_name (output);
+	len = name ? strlen (name) : 0;
+
 	if (container) {
-		cnlen = stream_name_len (container);
+		cnlen = stream_name_len (GSF_OUTFILE_ZIP (container));
 		if (cnlen > 0)
 			len += 1 + cnlen;
 	}
@@ -197,15 +232,23 @@ stream_name_len (GsfOutfile *outfile)
 }
 
 static void
-stream_name_write_to_buf (GsfOutfile *outfile, char *buf, int buflen)
+stream_name_write_to_buf (GsfOutfileZip *zip, char *buf, int buflen)
 {
-	GsfOutput  *output = GSF_OUTPUT (outfile);
-	GsfOutfile *container = gsf_output_container (output);
-	char const *name = gsf_output_name (output);
+	GsfOutput  *output;
+	GsfOutfile *container;
+	char const *name;
 	int len = 0;
 
+	if (zip == zip->root)
+		return;
+
+	output = GSF_OUTPUT (zip);
+	container = gsf_output_container (output);
+	name = gsf_output_name (output);
+
 	if (container) {
-		stream_name_write_to_buf (container, buf, buflen);
+		stream_name_write_to_buf (GSF_OUTFILE_ZIP (container),
+					  buf, buflen);
 		len = strlen (buf);
 		if (len > 0) {
 			buf[len++] = '/';
@@ -214,17 +257,16 @@ stream_name_write_to_buf (GsfOutfile *outfile, char *buf, int buflen)
 	}
 	if (name)
 		strncpy (buf + len, name, buflen - len);
-
 }
 
 static char *
-stream_name_build (GsfOutfile *outfile)
+stream_name_build (GsfOutfileZip *zip)
 {
-	int namelen = stream_name_len (outfile);
+	int namelen = stream_name_len (zip);
 	char *name = g_malloc (namelen + 1);
 
 	name[0] = '\0';
-	stream_name_write_to_buf (outfile, name, namelen + 1);
+	stream_name_write_to_buf (zip, name, namelen + 1);
 
 	return name;
 }
@@ -248,10 +290,9 @@ zip_time_make (time_t *t)
 static ZipDirent*
 zip_dirent_new_out (GsfOutfileZip *zip)
 {
-	GsfOutfile *outfile = GSF_OUTFILE (zip);
-	ZipDirent *dirent = g_new (ZipDirent, 1);
+	ZipDirent *dirent;
 	time_t t = time (NULL);
-	char *name = stream_name_build (outfile);
+	char *name = stream_name_build (zip);
 
 	if (!name)
 		return NULL;
@@ -301,16 +342,25 @@ zip_init_write (GsfOutput *output)
 	GsfOutfileZip *zip = GSF_OUTFILE_ZIP (output);
 	ZipDirent *dirent;
 	int      ret;
-	
+
+	if (zip->root->writing) {
+		g_warning ("Already writing to another stream in archive");
+		return FALSE;
+	}
+
 	if (!gsf_output_wrap (zip->sink, output)) {
 		return FALSE;
 	}
 
 	dirent = zip_dirent_new_out (zip);
 	dirent->offset = gsf_output_tell (zip->sink);
+	if (zip->vdir->dirent)
+		g_warning ("Leak.");
+
 	zip->vdir->dirent = dirent;
 	zip_header_write (zip);
 	zip->writing = TRUE;
+	zip->root->writing = TRUE;
 	dirent->crc32 = crc32 (0L, Z_NULL, 0);
 	if (zip->compression_method == GSF_ZIP_DEFLATED) {
 		if (!zip->stream) {
@@ -322,7 +372,7 @@ zip_init_write (GsfOutput *output)
 		if (ret != Z_OK)
 			return FALSE;
 		if (!zip->buf) {
-			zip->buf_size = ZIP_BUF_SIZE; 
+			zip->buf_size = ZIP_BUF_SIZE;
 			zip->buf = g_malloc (zip->buf_size);
 		}
 		zip->stream->next_out  = zip->buf;
@@ -337,7 +387,7 @@ zip_output_block (GsfOutfileZip *zip)
 {
 	size_t num_bytes = zip->buf_size - zip->stream->avail_out;
 	ZipDirent *dirent = zip->vdir->dirent;
-	
+
 	if (!gsf_output_write (zip->sink, num_bytes, zip->buf)) {
 		return FALSE;
 	}
@@ -378,7 +428,7 @@ zip_ddesc_write (GsfOutfileZip *zip)
 		{ 'P', 'K', 0x07, 0x08 };
 	guint8 buf[16];
 	ZipDirent *dirent = zip->vdir->dirent;
-	
+
 	memcpy (buf, ddesc_signature, sizeof ddesc_signature);
 	GSF_LE_SET_GUINT32 (buf + 4, dirent->crc32);
 	GSF_LE_SET_GUINT32 (buf + 8, dirent->csize);
@@ -386,7 +436,7 @@ zip_ddesc_write (GsfOutfileZip *zip)
 	if (!gsf_output_write (zip->sink, sizeof buf, buf)) {
 		return FALSE;
 	}
-	
+
 	return TRUE;
 }
 
@@ -396,7 +446,6 @@ zip_header_write_sizes (GsfOutfileZip *zip)
 	guint8 hbuf[ZIP_HEADER_SIZE];
 	ZipDirent *dirent = zip->vdir->dirent;
 	gsf_off_t pos = gsf_output_tell (zip->sink);
-	gsf_off_t hpos = dirent->offset;
 
 	if (!gsf_output_seek (zip->sink, dirent->offset + ZIP_HEADER_CRC,
 			      G_SEEK_SET))
@@ -416,7 +465,7 @@ zip_header_write_sizes (GsfOutfileZip *zip)
 static gboolean
 zip_close_stream (GsfOutput *output)
 {
-	GsfOutfileZip *zip = GSF_OUTFILE_ZIP (output);	
+	GsfOutfileZip *zip = GSF_OUTFILE_ZIP (output);
 
 	if (!zip->writing)
 		if (!zip_init_write (output))
@@ -425,14 +474,15 @@ zip_close_stream (GsfOutput *output)
 	if (zip->compression_method == GSF_ZIP_DEFLATED) {
 		if (!zip_flush (zip))
 			return FALSE;
-		
+
 		if (!zip_ddesc_write (zip)) /* Write data descriptor */
 			return FALSE;
 	} else {
 		if (!zip_header_write_sizes (zip)) /* Write crc, sizes */
 			return FALSE;
 	}
-	
+	zip->root->writing = FALSE;
+
 	return gsf_output_unwrap (zip->sink, output);
 }
 
@@ -462,7 +512,7 @@ gsf_outfile_zip_write (GsfOutput *output,
 	GsfOutfileZip *zip = GSF_OUTFILE_ZIP (output);
 	ZipDirent *dirent;
 	int ret;
-	
+
 	g_return_val_if_fail (zip && zip->vdir, FALSE);
 	g_return_val_if_fail (!zip->vdir->is_directory, FALSE);
 	g_return_val_if_fail (data, FALSE);
@@ -475,7 +525,7 @@ gsf_outfile_zip_write (GsfOutput *output,
 	if (zip->compression_method == GSF_ZIP_DEFLATED) {
 		zip->stream->next_in  = (unsigned char *) data;
 		zip->stream->avail_in = num_bytes;
-		
+
 		while (zip->stream->avail_in > 0) {
 			if (zip->stream->avail_out == 0) {
 				if (!zip_output_block (zip))
@@ -500,9 +550,10 @@ static void
 root_register_child (GsfOutfileZip *root, GsfOutfileZip *child)
 {
 	child->root = root;
-	g_object_ref (G_OBJECT (child));
-	if (!child->vdir->is_directory)
+	if (!child->vdir->is_directory) {
+		g_object_ref (G_OBJECT (child));
 		g_ptr_array_add (root->root_order, child);
+	}
 }
 
 static GsfOutput *
@@ -524,7 +575,7 @@ gsf_outfile_zip_new_child (GsfOutfile *parent, char const *name,
 	gsf_output_set_container (GSF_OUTPUT (child), parent);
 	vdir_add_child (zip_parent->vdir, child->vdir);
 	root_register_child (zip_parent->root, child);
-	
+
 	return GSF_OUTPUT (child);
 }
 
@@ -575,7 +626,7 @@ GSF_CLASS (GsfOutfileZip, gsf_outfile_zip,
  * Returns : the new zip file handler
  **/
 GsfOutfile *
-gsf_outfile_zip_new (GsfOutput *sink, GError **err)
+gsf_outfile_zip_new (GsfOutput *sink, G_GNUC_UNUSED GError **err)
 {
 	GsfOutfileZip *zip;
 
@@ -584,7 +635,7 @@ gsf_outfile_zip_new (GsfOutput *sink, GError **err)
 	zip = g_object_new (GSF_OUTFILE_ZIP_TYPE, NULL);
 	g_object_ref (G_OBJECT (sink));
 	zip->sink = sink;
-	
+
 	zip->vdir = vdir_new ("", TRUE, NULL);
 	zip->root_order = g_ptr_array_new ();
 	zip->root = zip;
@@ -604,7 +655,7 @@ gsf_outfile_zip_set_compression_method (GsfOutfileZip *zip,
 
 	if (zip->writing || (zip->vdir && zip->vdir->is_directory))
 		return FALSE;
-	
+
 	switch (method) {
 	case GSF_ZIP_STORED:
 	case GSF_ZIP_DEFLATED:
