@@ -37,16 +37,21 @@ struct _GsfOutput {
 	GsfOutfile *container;
 	GError	   *err;
 	gboolean    is_closed;
+
+	char	   *printf_buf;
+	int	    printf_buf_size;
 };
 
 typedef struct {
 	GObjectClass g_object_class;
 
-	gboolean (*Close) (GsfOutput *output);
-	gboolean (*Seek)  (GsfOutput *output,
-			   gsf_off_t offset, GSeekType whence);
-	gboolean (*Write) (GsfOutput *output,
-			   size_t num_bytes, guint8 const *data);
+	gboolean (*Close)   (GsfOutput *output);
+	gboolean (*Seek)    (GsfOutput *output,
+			     gsf_off_t offset, GSeekType whence);
+	gboolean (*Write)   (GsfOutput *output,
+			     size_t num_bytes, guint8 const *data);
+	gboolean (*Vprintf) (GsfOutput *output,
+			     char const *format, va_list args);
 } GsfOutputClass;
 
 #define GSF_OUTPUT_CLASS(k)    (G_TYPE_CHECK_CLASS_CAST ((k), GSF_OUTPUT_TYPE, GsfOutputClass))
