@@ -67,74 +67,51 @@ follow_symlinks (char const *filename, GError **error)
 	followed_filename = g_strdup (filename);
 	link_count = 0;
 
-	while (link_count < GSF_MAX_LINK_LEVEL) {
-		struct stat st;
+	while (link_count++ < GSF_MAX_LINK_LEVEL) {
+		char linkname[GSF_MAX_PATH_LEN];
+		int len = readlink (followed_filename, linkname, GSF_MAX_PATH_LEN - 1);
 
-		if (lstat (followed_filename, &st) != 0)
-			/* We could not access the file, so perhaps it does not
-			 * exist.  Return this as a real name so that we can
-			 * attempt to create the file.
-			 */
-			return followed_filename;
+		if (len == -1) {
+			switch (errno) {
+			case EINVAL: /* Not a symlink.  */
+			case ENOSYS: /* Surely not a symlink.  */
+			case ENOENT: /* No such file.  */
+				return followed_filename;
 
-		if (S_ISLNK (st.st_mode)) {
-			gint len;
-			gchar linkname[GSF_MAX_PATH_LEN];
-
-			link_count++;
-
-			len = readlink (followed_filename, linkname, GSF_MAX_PATH_LEN - 1);
-
-			if (len == -1) {
-				char *utf8name = gsf_filename_to_utf8 (followed_filename, FALSE);
-				g_set_error (error, gsf_output_error_id (), errno,
-					     "Could not read symbolic link information "
-					       "for %s", utf8name);
-				g_free (utf8name);
+			default:
+				if (error)
+					*error = g_error_new (gsf_output_error_id (), errno,
+							      g_strerror (errno));
 				g_free (followed_filename);
 				return NULL;
 			}
+		}
 
-			linkname[len] = '\0';
+		linkname[len] = '\0';
 
-			/* If the linkname is not an absolute path name, append
+		if (g_path_is_absolute (linkname)) {
+			g_free (followed_filename);
+			followed_filename = g_strdup (linkname);
+		} else {
+			/*
+			 * If the linkname is not an absolute path name, append
 			 * it to the directory name of the followed filename.  E.g.
 			 * we may have /foo/bar/baz.lnk -> eek.txt, which really
 			 * is /foo/bar/eek.txt.
 			 */
 
-			if (linkname[0] != G_DIR_SEPARATOR)
-			{
-				gchar *slashpos;
-				gchar *tmp;
+			char *dir = g_path_get_dirname (followed_filename);
 
-				slashpos = strrchr (followed_filename, G_DIR_SEPARATOR);
-
-				if (slashpos)
-					*slashpos = '\0';
-				else
-				{
-					tmp = g_strconcat ("./", followed_filename, NULL);
-					g_free (followed_filename);
-					followed_filename = tmp;
-				}
-
-				tmp = g_build_filename (followed_filename, linkname, NULL);
-				g_free (followed_filename);
-				followed_filename = tmp;
-			}
-			else
-			{
-				g_free (followed_filename);
-				followed_filename = g_strdup (linkname);
-			}
-		} else
-			return followed_filename;
+			g_free (followed_filename);
+			followed_filename = g_build_filename (dir, linkname, NULL);
+			g_free (dir);
+		}
 	}
 
 	/* Too many symlinks */
-	g_set_error (error, gsf_output_error_id (), ELOOP,
-		     "The file has too many symbolic links.");
+	if (error)
+		*error = g_error_new (gsf_output_error_id (), ELOOP,
+				      g_strerror (ELOOP));
 
 	return NULL;
 }
@@ -165,10 +142,9 @@ gsf_output_stdio_new (char const *filename, GError **err)
 {
 	GsfOutputStdio *stdio;
 	FILE *file = NULL;
-	char *slashpos, *dirname, *temp_filename = NULL;
+	char *dirname, *temp_filename = NULL;
 	char *real_filename = follow_symlinks (filename, err);
 	int fd;
-	gboolean create_backup_copy = TRUE;
 	mode_t saved_umask;
 	struct stat st;
 	
@@ -176,27 +152,34 @@ gsf_output_stdio_new (char const *filename, GError **err)
 		return NULL;
 
 	/* Get the directory in which the real filename lives */
-	slashpos = strrchr (real_filename, G_DIR_SEPARATOR);
-	if (slashpos) {
-		dirname = g_strdup (real_filename);
-		dirname[slashpos - real_filename + 1] = '\0';
-	} else
-		dirname = g_strdup (".");
+	dirname = g_path_get_dirname (real_filename);
+
+	if (stat (real_filename, &st) == 0) {
+		/* FIXME: use eaccess if available.  */
+		/* FIXME? Race conditions en masse.  */
+		if (access (real_filename, W_OK) != 0) {
+			if (err != NULL)
+				*err = g_error_new (gsf_output_error_id (), errno,
+						    g_strerror (errno));
+			goto failure;
+		}
+	} else {
+		/*
+		 * File does not exist.  Compute the permissions and uid/gid
+		 * that we will use for the newly-created file.
+		 */
 
-	/* If there is not an existing file with that name, compute the
-	 * permissions and uid/gid that we will use for the newly-created file.
-	 */
-	if (stat (real_filename, &st) != 0) {
 		struct stat dir_st;
 		int result;
 
-		/* File does not exist? */
-		create_backup_copy = FALSE;
-
 		/* Use default permissions */
 		st.st_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
 		st.st_uid = getuid ();
 
+		/* Used to set create_backup_copy = FALSE, but
+		   creating backup copies on this level is a 
+		   horrible mistake. -DAL- */
+
 		result = stat (dirname, &dir_st);
 
 		if (result == 0 && (dir_st.st_mode & S_ISGID))
@@ -210,26 +193,25 @@ gsf_output_stdio_new (char const *filename, GError **err)
 	 */
 	temp_filename = g_build_filename (dirname, ".gsf-save-XXXXXX", NULL);
 	g_free (dirname);
+	/* Oh, joy.  What about threads?  --MW */
 	saved_umask = umask (0077);
 	fd = g_mkstemp (temp_filename); /* this modifies temp_filename to the used name */
 	umask (saved_umask);
 
-	if (fd < 0 || NULL == (file = fdopen (fd, "w"))) {
-		if (err != NULL) {
-			char *utf8name = gsf_filename_to_utf8 (temp_filename, FALSE);
+	if (fd < 0 || NULL == (file = fdopen (fd, "wb"))) {
+		if (err != NULL)
 			*err = g_error_new (gsf_output_error_id (), errno,
-				"%s: %s", utf8name, g_strerror (errno));
-			g_free (utf8name);
-		}
+					    g_strerror (errno));
 		goto failure;
 	}
 
 	stdio = g_object_new (GSF_OUTPUT_STDIO_TYPE, NULL);
 	stdio->file = file;
 	stdio->st = st;
-	stdio->create_backup_copy = create_backup_copy;
+	stdio->create_backup_copy = FALSE;
 	stdio->real_filename = real_filename;
 	stdio->temp_filename = temp_filename;
+	gsf_output_set_name (GSF_OUTPUT (stdio), filename);
 
 	return GSF_OUTPUT (stdio);
 
@@ -284,9 +266,17 @@ gsf_output_stdio_close (GsfOutput *output)
 	} else {
 		/* Restore permissions.  There is not much error checking we
 		 * can do here, I'm afraid.  The final data is saved anyways.
+		 * Note the order: mode, uid+gid, gid, uid, mode.
 		 */
 		chmod (stdio->real_filename, stdio->st.st_mode);
-		chown (stdio->real_filename, stdio->st.st_uid, stdio->st.st_gid);
+		if (chown (stdio->real_filename,
+			   stdio->st.st_uid,
+			   stdio->st.st_gid)) {
+			/* We cannot set both.  Maybe we can set one.  */
+			chown (stdio->real_filename, -1, stdio->st.st_gid);
+			chown (stdio->real_filename, stdio->st.st_uid, -1);
+		}
+		chmod (stdio->real_filename, stdio->st.st_mode);
 	}
 
 	g_free (backup_filename);
@@ -359,6 +349,15 @@ gsf_output_stdio_write (GsfOutput *output,
 	return res == num_bytes;
 }
 
+static gboolean
+gsf_output_stdio_vprintf (GsfOutput *output, char const *format, va_list args)
+{
+	GsfOutputStdio *stdio = (GsfOutputStdio *)output;
+	int res = vfprintf (stdio->file, format, args);
+
+	return (res >= 0);
+}
+
 static void
 gsf_output_stdio_init (GObject *obj)
 {
@@ -377,6 +376,7 @@ gsf_output_stdio_class_init (GObjectClass *gobject_class)
 	output_class->Close	= gsf_output_stdio_close;
 	output_class->Seek	= gsf_output_stdio_seek;
 	output_class->Write	= gsf_output_stdio_write;
+	output_class->Vprintf	= gsf_output_stdio_vprintf;
 }
 
 GSF_CLASS (GsfOutputStdio, gsf_output_stdio,
